#pragma once

#include<stddef.h>
#include<stdint.h>
#include<cglm/mat4.h>
#include<string.h>
#include<ode/ode.h>
#include<cglm/quat.h>
#include"k3.h"

#define GAME_TPS 50

#define CATEGORY_STATIC 1
#define CATEGORY_RAY 2
#define CATEGORY_ENTITY 4
#define CATEGORY_GHOST 8

#define TRIGGER_INVALID 0

#define MODELNAME_LENGTH 12

struct CRender {
	uint16_t entity;
	char mdl[MODELNAME_LENGTH];
	
	struct k3Mdl *cache;
	
	vec4 pos;
	versor rot;
	
	vec4 posLast;
	versor rotLast;
};

struct TrimeshData {
	void *vdata;
	void *idata;
	dTriMeshDataID trid;
};

#define CPHYSICS_TRIMESH 0
#define CPHYSICS_CAPSULE 1
#define CPHYSICS_BOX 2
#define CPHYSICS_SPHERE 3
#define CPHYSICS_STATIC 0
#define CPHYSICS_KINEMATIC 1
#define CPHYSICS_DYNAMIC 2
#define CPHYSICS_GHOST 128
struct CPhysics {
	uint16_t entity;
	dGeomID geom, geom2;
	uint16_t trigger;
	
	uint8_t type;
	uint8_t dynamics;
	union {
		struct {
			char name[MODELNAME_LENGTH];
			struct TrimeshData *cache;
		} trimesh;
		struct {
			float length;
			float radius;
		} capsule;
		struct {
			float w;
			float h;
			float l;
		} box;
		struct {
			float radius;
		} sphere;
	};
	
	vec3 start;
	vec3 vel;
	vec4 rot;
	
	float speed;
	
	float mass;
	float friction;
	
	uint32_t collide;
};

struct CMovement {
	uint16_t entity;
	vec3 dir;
	float pointing;
	char jump, canJump, isJumping;
	vec3 groundNormal;
	float groundDepth;
	uint16_t holding;
};

struct LocalRay {
	vec3 pos;
	vec3 dir;
	uint16_t ignore;
	
	vec3 out;
	uint16_t hit;
	float depth;
	
	float maxlen;
};

struct CPlayerCtrl {
	uint16_t entity;
	float yaw;
	float pitch;
	uint32_t keys;
	
	struct LocalRay holdray;
};

struct CBoned {
	uint16_t entity;
	
	size_t size;
	struct k3AnimationBone *bones;
	
	struct {
		uint16_t id;
		struct k3Animator cache;
		bool standard;
	} anim;
};

#define CONVEYOR_TYPE_LOOP 0
#define CONVEYOR_TYPE_BINARY 1
struct Conveyor {
	uint16_t entity;
	int16_t active;
	uint8_t type;
	float speed;
	float position;
	size_t pointCount;
	vec3 *points;
	
	uint16_t trigger;
	
	uint8_t effectivelyActive;
	uint8_t effectivelyActiveLast;
};

#define TRIGGER_EV_IN 0
#define TRIGGER_EV_CONTINUOUS 1
#define TRIGGER_EV_OUT 2
typedef int(*TriggerFunc)(size_t id, uint16_t ethis, uint16_t ethat, uint8_t event);

#define MAX_ENTS 128
extern struct Game {
	size_t entityCount;
	struct {
		struct CRender render[MAX_ENTS];
		size_t renderCount;
		
		struct CPhysics physics[MAX_ENTS];
		size_t physicsCount;
		
		struct CMovement movement[MAX_ENTS];
		size_t movementCount;
		
		struct CPlayerCtrl playerctrl[MAX_ENTS];
		size_t playerctrlCount;
		
		struct CBoned boned[MAX_ENTS];
		size_t bonedCount;
		
		uint64_t freeIDs[(MAX_ENTS + 63) / 64];
	} entities;
	
	int isMultiplayer;
	int isAuthority;
	int reconciliate;
	
	size_t frame;
	
#define MAX_CONVEYORS 4096
	size_t conveyorCount;
	struct Conveyor **conveyors;
	
#define MAX_TRIGGERS 4096
	size_t triggerCount;
	TriggerFunc *triggers;
	
	dWorldID phys;
	dSpaceID space;
	dJointGroupID contactgroup;
	
	struct TrimeshData *statikRes;
	dGeomID statik;
	
	uint16_t spectated;
	uint16_t controlled;
} Game;

void game_init();

void game_raycast(struct LocalRay *rays, size_t count);
void game_update();

void game_setphys(struct TrimeshData*);

#define ENT_ID_INVALID 65535
static inline size_t game_nextid() {
	for(size_t i = 0; i < (MAX_ENTS + 63) / 64; i++) {
		for(uint8_t b = 0; b < 64; b++) {
			if((Game.entities.freeIDs[i] & (1ULL << b)) == 0) {
				return i * 8 + b;
			}
		}
	}
	return ENT_ID_INVALID;
}

static inline void game_claimentid(size_t i) {
	Game.entities.freeIDs[i / 64] |= 1UL << (i % 64);
}

static inline int game_componentcomparator(const void *a, const void *b) {
	return *(const uint16_t*) a - *(const uint16_t*) b;
}

#define game_addcomponent(a, c) game_addcomponent_((uint8_t*) (Game.entities.a), (uint8_t*) (c), sizeof(*(Game.entities.a)), &(Game.entities.a##Count))
static inline void game_addcomponent_(uint8_t *array, uint8_t *component, size_t componentSize, size_t *componentCount) {
	memcpy(&array[*componentCount * componentSize], component, componentSize);
	
	(*componentCount)++;
	
	qsort(array, *componentCount, componentSize, game_componentcomparator);
}

#define game_getcomponent(eid, arr) ((typeof(&Game.entities.arr[0]))game_getcomponent_((eid), (Game.entities.arr), sizeof(*(Game.entities.arr)), (Game.entities.arr##Count)))
static inline void *game_getcomponent_(uint16_t eid, void *array, size_t componentSize, size_t componentCount) {
	return bsearch(&eid, array, componentCount, componentSize, game_componentcomparator);
}

#define game_ensurecomponent(eid, arr) ((typeof(&Game.entities.arr[0]))game_ensurecomponent_((eid), (Game.entities.arr), sizeof(*(Game.entities.arr)), &(Game.entities.arr##Count)))
static inline void *game_ensurecomponent_(uint16_t eid, void *array, size_t componentSize, size_t *componentCount) {
	void *ptr = game_getcomponent_(eid, array, componentSize, *componentCount);
	if(!ptr) {
		void *data = alloca(componentSize);
		memset(data, 0, componentSize);
		*(uint16_t*) data = eid;
		game_addcomponent_(array, data, componentSize, componentCount);
		ptr = game_getcomponent_(eid, array, componentSize, *componentCount);
	}
	return ptr;
}

#define game_killcomponent_ptr(c, arr) do {									\
	size_t idx = c - Game.entities.arr;										\
	memmove(c, c + 1, sizeof(*c) * (Game.entities.arr##Count - idx - 1));	\
	Game.entities.arr##Count--;												\
} while(0);

#define game_killcomponent(eid, arr) do {							\
	typeof(&Game.entities.arr[0]) c = game_getcomponent(eid, arr);	\
	if(c) {															\
		game_killcomponent_ptr(c, arr);								\
	}																\
} while(0);

// An artifact of using a third-party physics engine
void game_synccphysics();

void game_cleanup();

void game_killentity(uint16_t eid);
